import os
import time
import uuid
from typing import Tuple

from podonos.core.base import log


def generate_random_name():
    current_time_milliseconds = int(time.time() * 1000)
    random_uuid = uuid.uuid4()
    return f"{current_time_milliseconds}-{random_uuid}"


def generate_random_group_name():
    current_time_milliseconds = int(time.time() * 1000)
    random_uuid = uuid.uuid4()
    return f"{current_time_milliseconds}_{random_uuid}"


def process_paths_to_posix(original_path: str, remote_object_path: str) -> Tuple[str, str]:
    """Convert paths to POSIX style

    Args:
        original_path: Original file path
        remote_object_path: Remote object path

    Returns:
        Tuple of processed original and remote paths
    """
    return (original_path.replace("\\", "/"), remote_object_path.replace("\\", "/"))


def get_content_type_by_filename(path: str) -> str:
    log.check_notnone(path)
    log.check_ne(path, "")

    _, ext = os.path.splitext(path)
    if ext == ".wav":
        return "audio/wav"
    elif ext == ".mp3":
        return "audio/mpeg"
    elif ext == ".flac":
        return "audio/flac"
    elif ext == ".json":
        return "application/json"
    return "application/octet-stream"



from __future__ import annotations
import atexit
import datetime
import queue
import threading
import time

from concurrent.futures import ThreadPoolExecutor
from threading import Event
from tqdm import tqdm
from typing import Optional, Any, Dict, Tuple

try:
    from typing import Protocol  # Python 3.8+: available in typing
except Exception:  # pragma: no cover
    from typing_extensions import Protocol  # type: ignore

from podonos.core.base import *
from podonos.service.evaluation_service import EvaluationService
from podonos.common.validator import Rules, validate_args


class UploadManager:
    """Concurrent file upload manager.
    Internally creates multiple threads, and manages the uploading status.
    """

    class UploadQueue(Protocol):
        def put(self, item: Tuple[str, str, str], block: bool = True, timeout: Optional[float] = None) -> None: ...
        def get(self, block: bool = True, timeout: Optional[float] = None) -> Tuple[str, str, str]: ...
        def empty(self) -> bool: ...
        def task_done(self) -> None: ...
        def join(self) -> None: ...

    # File path queue
    # TODO: use a file queue.
    _queue: Optional[UploadQueue] = None
    # Total number of files added to the uploading queue.
    _total_files: int = 0
    _total_uploaded: int = 0

    _pbar: Optional[Any] = None
    # Event to all the uploader threads
    _worker_event: Optional[Event] = None
    # Master daemon thread. Alive until the manager closes.
    _daemon_thread: Optional[threading.Thread] = None
    # Evaluation service
    _evaluation_service: EvaluationService
    # Manager status. True if the manager is ready.
    _status: bool = False
    # Maximum number of uploader worker threads
    _max_workers: int = 1
    #
    _upload_start: Optional[Dict[str, str]] = None
    _upload_finish: Optional[Dict[str, str]] = None

    def get_upload_time(self) -> Tuple[Dict[str, str], Dict[str, str]]:
        if not self._upload_start or not self._upload_finish:
            raise ValueError("Upload Fail")

        return self._upload_start, self._upload_finish

    @validate_args(evaluation_service=Rules.instance_of(EvaluationService), max_workers=Rules.positive_not_none)
    def __init__(
        self,
        evaluation_service: EvaluationService,
        max_workers: int,
    ) -> None:
        self._upload_start = dict()
        self._upload_finish = dict()
        self._evaluation_service = evaluation_service
        self._queue = queue.Queue()
        self._total_files = 0
        self._max_workers = max_workers
        self._worker_event = Event()
        self._daemon_thread = threading.Thread(target=self._uploader_daemon, daemon=True)
        self._daemon_thread.start()
        self._status = True

        atexit.register(self.wait_and_close)

    def _uploader_daemon(self) -> None:
        log.debug(f"Uploader daemon is running with {self._max_workers} workers")
        with ThreadPoolExecutor(max_workers=self._max_workers) as executor:
            for index in range(self._max_workers):
                executor.submit(self._upload_worker, index, self._worker_event)  # type: ignore
        log.debug(f"Uploader daemon is shutting down")
        executor.shutdown(wait=True)

    @validate_args(index=Rules.int_not_none, worker_event=Rules.instance_of(Event))
    def _upload_worker(self, index: int, worker_event: Event) -> None:
        # Individual worker for uploading files. The upload manager creates multiple threads for each of this worker.
        if not (
            self._queue is not None
            and self._worker_event is not None
            and self._daemon_thread is not None
            and self._evaluation_service is not None  # type: ignore
            and self._upload_start is not None
            and self._upload_finish is not None
        ):
            raise ValueError("Upload Manager is not initialized")

        log.debug(f"Worker is {index} ready")
        while True:
            if not self._queue.empty():
                item: Tuple[str, str, str] = self._queue.get()
                evaluation_id = item[0]
                remote_object_name = item[1]
                path = item[2]

                log.debug(f"Worker {index} presigned url request")
                presigned_url = self._evaluation_service.get_presigned_url(
                    evaluation_id,
                    remote_object_name,
                )
                log.debug(f"Worker {index} presigned url obtained")

                log.debug(f"Worker {index} uploading {path}")
                # Timestamp in ISO 8601.
                upload_start_at = datetime.datetime.now().astimezone().isoformat(timespec="milliseconds")
                self._evaluation_service.upload_evaluation_file(presigned_url, path)
                upload_finish_at = datetime.datetime.now().astimezone().isoformat(timespec="milliseconds")
                log.debug(f"Worker {index} finished uploading {item}")

                self._upload_start[remote_object_name] = upload_start_at
                self._upload_finish[remote_object_name] = upload_finish_at
                self._queue.task_done()
                self._total_uploaded += 1
                log.debug(f"Worker {index} total_uploaded: {self._total_uploaded}")
                if self._pbar:
                    self._pbar.update(1)

            time.sleep(0.1)
            if worker_event.is_set():
                log.debug(f"Worker {index} is done")
                return

    def add_file_to_queue(self, evaluation_id: str, remote_object_name: str, path: str) -> None:
        if not (
            self._queue is not None
            and self._worker_event is not None
            and self._daemon_thread is not None
            and self._evaluation_service is not None  # type: ignore
            and self._upload_start is not None
            and self._upload_finish is not None
        ):
            raise ValueError("Upload Manager is not initialized")

        log.debug(f"Added: {path}")
        self._queue.put((evaluation_id, remote_object_name, path))
        self._total_files += 1

    def wait_and_close(self) -> bool:
        if not self._status:
            return False

        if not (self._queue is not None and self._worker_event is not None and self._daemon_thread is not None):
            raise ValueError("Upload Manager is not initialized")
        log.debug(f"total_files: {self._total_files}")
        self._pbar = tqdm(total=self._total_files, dynamic_ncols=True)
        self._pbar.update(self._total_uploaded)

        # Block until all tasks are done.
        log.debug("Queue join")
        self._queue.join()

        # Signal all the workers to exit.
        log.debug("Set exit event to workers")
        self._worker_event.set()
        # Shutdown the executor.
        log.debug("Shutdown uploader daemon")
        self._daemon_thread.join()

        self._pbar.close()
        log.info("All upload work complete.")
        self._status = False
        return True

    def _check_if_initialize(self) -> bool:
        return (
            self._queue is not None
            and self._worker_event is not None
            and self._daemon_thread is not None
            and self._evaluation_service is not None  # type: ignore
            and self._upload_start is not None
            and self._upload_finish is not None
        )



from typing import Dict, List, Optional


from podonos.common.enum import EvalType
from podonos.core.api import APIClient
from podonos.core.base import log
from podonos.core.config import EvalConfig
from podonos.common.validator import Rules, validate_args
from podonos.core.file import File, Audio, AudioGroup, FileTransformer, FileValidator
from podonos.core.upload_manager import UploadManager
from podonos.entity.evaluation import EvaluationEntity
from podonos.service.evaluation_service import EvaluationService


class Evaluator:
    """Base class for all evaluators."""

    _api_client: APIClient
    _eval_config: EvalConfig
    _evaluation: Optional[EvaluationEntity] = None
    _evaluation_service: EvaluationService
    _file_transformer: FileTransformer
    _file_validator: FileValidator
    _supported_eval_types: List[EvalType]
    _initialized: bool = False
    _upload_manager: Optional[UploadManager] = None  # Upload manager. Lazy initialization when used for saving resources.
    _ordered_file_groups: List[AudioGroup]  # ordered evaluation files groups

    def __init__(self, api_client: APIClient, eval_config: EvalConfig, supported_eval_types: List[EvalType]):
        """Initialize the evaluator.

        Args:
            api_client: API client for making requests
            eval_config: Optional evaluation configuration
        """
        self._initialized = True
        self._validate_initialization(api_client, eval_config, supported_eval_types)
        self._initialize_attributes(api_client, eval_config, supported_eval_types)

    @validate_args(api_client=Rules.instance_of(APIClient), eval_config=Rules.instance_of(EvalConfig), supported_eval_types=Rules.list_not_none)
    def _validate_initialization(self, api_client: APIClient, eval_config: EvalConfig, supported_eval_types: List[EvalType]) -> None:
        """Validate the initialization parameters.

        Args:
            api_client: API client to validate
            eval_config: Evaluation configuration to validate

        Raises:
            ValueError: If api_client is not initialized
            ValueError: If eval_config is not initialized
        """
        if not api_client:
            raise ValueError("api_client is not initialized.")
        if not eval_config:
            raise ValueError("eval_config is not initialized.")
        if eval_config.eval_type not in supported_eval_types:
            raise ValueError("Not supported evaluation type")

    @validate_args(api_client=Rules.instance_of(APIClient), eval_config=Rules.instance_of(EvalConfig), supported_eval_types=Rules.list_not_none)
    def _initialize_attributes(self, api_client: APIClient, eval_config: EvalConfig, supported_eval_types: List[EvalType]) -> None:
        """Initialize class attributes.

        Args:
            api_client: API client for making requests
            eval_config: Evaluation configuration
        """
        self._api_client = api_client
        self._eval_config = eval_config
        self._evaluation_service = EvaluationService(api_client)
        self._evaluation = self._set_evaluation(eval_config)
        self._file_transformer = FileTransformer(eval_config)
        self._file_validator = FileValidator(eval_config)
        self._supported_eval_types = supported_eval_types
        self._ordered_file_groups = []
        self._upload_manager = None

    @validate_args(method_name=Rules.str_non_empty)
    def _validate_eval_type(self, method_name: str) -> None:
        """Validate if the evaluation type is supported for the given method.

        Args:
            method_name: Name of the method being validated ('add_file' or 'add_files')

        Raises:
            ValueError: If evaluation type is not supported for the method
        """
        if method_name == "add_file":
            supported_types = [EvalType.NMOS, EvalType.QMOS, EvalType.P808, EvalType.CUSTOM_SINGLE]
            error_msg = f"The '{method_name}' is only supported for single file evaluation types: " f"{supported_types}"
        else:  # add_files
            supported_types = [EvalType.CMOS, EvalType.DMOS, EvalType.PREF, EvalType.SMOS, EvalType.CSMOS, EvalType.CUSTOM_DOUBLE]
            error_msg = f"The '{method_name}' is only supported for comparison evaluation types: " f"{supported_types}"

        if self._eval_config.eval_type not in supported_types:
            raise ValueError(error_msg)

    def get_evaluation_id(self) -> str:
        """Get the evaluation ID.

        Returns:
            str: Evaluation ID

        Raises:
            AssertionError: If evaluation is not initialized
        """
        assert self._evaluation, "Evaluation not initialized"
        return self._evaluation.id

    def close(self) -> Dict[str, str]:
        """Close the evaluation session and upload results.

        Returns:
            Dict[str, str]: Status of the operation

        Raises:
            ValueError: If session is not initialized or upload manager is not defined
        """
        self._validate_close()
        self._wait_for_uploads()
        self._process_audio_files()
        self._upload_session_json()
        self._cleanup()
        return {"status": "ok"}

    @validate_args(file=Rules.instance_of(File))
    def add_file(self, file: File) -> None:
        """Add new file for speech evaluation.
        The file may be either in {wav, mp3} format. The file will be securely uploaded to
        Podonos service system.

        Args:
            file: File object including the path, the model tag, the other tags, and the script.

        Example:
        If you want to evaluate each audio file separately (e.g., Naturalness MOS):
            add_file(file=File(path='./test.wav', model_tag='my_new_model1', tags=['male', 'generated'],
                               script='hello there'))

        Returns: None

        Raises:
            ValueError: if this function is called before calling init()
            FileNotFoundError: if a given file is not found.
        """
        if not self._initialized:
            raise ValueError("Try to add file once the evaluator is closed.")

        self._validate_eval_type("add_file")

        file = self._file_validator.validate_file(file)
        audio_group = self._file_transformer.transform_into_audio_group([file])
        self._ordered_file_groups.append(audio_group)
        self._upload_one_file(
            evaluation_id=self.get_evaluation_id(), remote_object_name=audio_group.audios[0].remote_object_name, path=audio_group.audios[0].path
        )

    @validate_args(file0=Rules.instance_of(File), file1=Rules.instance_of(File), file2=Rules.optional_instance_of(File))
    def add_files(self, file0: File, file1: File, file2: Optional[File] = None) -> None:
        """Add two files for speech evaluation. The files will be securely uploaded to Podonos service system.

        The order of files is maintained based on evaluation type:
        - PREF, CUSTOM_DOUBLE: Files are ordered stimulus
        - SMOS: Files are unordered stimulus
        - CMOS, DMOS: One file must be reference, one must be stimulus
        - CSMOS: One file must be reference, two must be stimulus

        Args:
            file0: First audio file
            file1: Second audio file
            file2: Third audio file

        Example:
        If you want to evaluate audio files together (e.g., Comparative MOS):
            f0 = File(path="/path/to/generated.wav", model_tag='my_new_model1', tags=['male', 'english'], is_ref=True)
            f1 = File(path="/path/to/original.wav", model_tag='my_new_model2', tags=['male', 'english', 'param1'])
            add_files(file0=f0, file1=f1)

        If you want to evaluate two stimuli with a reference:
            ref = File(path="/path/to/reference.wav", model_tag='my_new_model3', tags=['male', 'english'], is_ref=True)
            add_files(file0=f0, file1=f1, file2=ref)

        Returns: None

        Raises:
            ValueError: If evaluator not initialized or invalid file configuration
        """
        if not self._initialized:
            raise ValueError("Evaluator is not initialized")

        self._validate_eval_type("add_files")

        files = self._file_validator.validate_files([file0, file1, file2])
        audio_group = self._file_transformer.transform_into_audio_group(files)
        self._ordered_file_groups.append(audio_group)
        for audio in audio_group.audios:
            self._upload_one_file(evaluation_id=self.get_evaluation_id(), remote_object_name=audio.remote_object_name, path=audio.path)

    def _validate_close(self) -> None:
        """Validate the state before closing.

        Raises:
            ValueError: If session is not properly initialized
        """
        if not self._initialized or not self._eval_config:
            raise ValueError("No evaluation session is open.")

    def _wait_for_uploads(self) -> None:
        """Wait for all file uploads to complete."""
        log.debug("Wait until the upload manager shuts down all the upload workers")
        assert self._upload_manager and self._upload_manager.wait_and_close()

    def _process_audio_files(self) -> None:
        """Process and upload audio files metadata."""
        log.info("Uploading the final pieces...")
        audios = [audio for group in self._ordered_file_groups for audio in group.audios]
        for i in range(0, len(audios), 500):
            self._evaluation_service.create_evaluation_files(self.get_evaluation_id(), audios[i : i + 500])

        self._process_upload_times()

    def _process_upload_times(self) -> None:
        """Process and store upload times for audio files."""
        if not self._upload_manager:
            return

        upload_start, upload_finish = self._upload_manager.get_upload_time()
        for group in self._ordered_file_groups:
            for audio in group.audios:
                self._update_audio_upload_times(audio, upload_start, upload_finish)

    @validate_args(audio=Rules.instance_of(Audio), upload_start=Rules.dict_not_none, upload_finish=Rules.dict_not_none)
    def _update_audio_upload_times(self, audio: Audio, upload_start: Dict[str, str], upload_finish: Dict[str, str]) -> None:
        """Update upload times for a single audio file.

        Args:
            audio: Audio object to update
            upload_start: Dictionary of upload start times
            upload_finish: Dictionary of upload finish times
        """
        remote_object_name = audio.remote_object_name
        upload_start_at = upload_start[remote_object_name]
        upload_finish_at = upload_finish[remote_object_name]
        audio.set_upload_at(upload_start_at, upload_finish_at)

    def _upload_session_json(self) -> None:
        """Upload the session JSON data."""
        self._evaluation_service.upload_session_json(self.get_evaluation_id(), self._eval_config, self._ordered_file_groups)

    def _cleanup(self) -> None:
        """Clean up the evaluation session."""
        self._initialized = False
        self._ordered_file_groups = []

    @validate_args(eval_config=Rules.instance_of(EvalConfig))
    def _set_evaluation(self, eval_config: EvalConfig) -> EvaluationEntity:
        if eval_config.eval_template_id:
            return self._evaluation_service.create_from_template(eval_config)
        return self._evaluation_service.create(eval_config)

    @validate_args(evaluation_id=Rules.str_non_empty, remote_object_name=Rules.str_non_empty, path=Rules.file_path_not_none)
    def _upload_one_file(
        self,
        evaluation_id: str,
        remote_object_name: str,
        path: str,
    ) -> None:
        """
        Start uploading one file to server.

        Args:
            evaluation_id: New evaluation's id.
            remote_object_name: Path to the remote file name.
            path: Path to the local file.
        Returns:
            None
        """
        # Get the presigned URL for one file
        log.debug(f"Adding to queue: {path}")
        if not self._eval_config:
            raise ValueError("No evaluation session is open.")

        # Lazy initialization of upload manager.
        if self._upload_manager is None:
            log.debug(f"max_upload_workers: {self._eval_config.max_upload_workers}")
            self._upload_manager = UploadManager(
                evaluation_service=self._evaluation_service,
                max_workers=self._eval_config.max_upload_workers,
            )

        if self._upload_manager:
            self._upload_manager.add_file_to_queue(evaluation_id, remote_object_name, path)
        return
